/*
 * mod_vhosterror.c - Apache 2.x module to produce error log prefixed with vhost
 * Copyright (C) 2011 Manuel Mausz <manuel@mausz.at>
 * Origin code from mod_globalerror (C) 2004 Ivan Ristic <ivanr@webkreator.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"
#include "http_connection.h"

#include "apr_strings.h"

module AP_MODULE_DECLARE_DATA vhosterror_module;

typedef struct
{
  const char *filename;
  int loglevel;
  apr_file_t *fd;
} server_config;

typedef struct
{
  char *t_name;
  int   t_val;
} TRANS;

static const TRANS priorities[] =
{
  { "emerg",   APLOG_EMERG },
  { "alert",   APLOG_ALERT },
  { "crit",    APLOG_CRIT },
  { "error",   APLOG_ERR },
  { "warn",    APLOG_WARNING },
  { "notice",  APLOG_NOTICE },
  { "info",    APLOG_INFO },
  { "debug",   APLOG_DEBUG },
  { NULL,      -1 },
};

static const char *cmd_vhost_error_log(cmd_parms *cmd, void *conf, const char *p1) {
  server_config *sconf = (server_config *)ap_get_module_config(
      cmd->server->module_config, &vhosterror_module);
  const char *err = ap_check_cmd_context(cmd,
      NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);

  if (err != NULL)
    return err;
  sconf->filename = p1;
  return NULL;
}

static const char *cmd_vhost_error_log_level(cmd_parms *cmd, void *conf, const char *p1)
{
  server_config *sconf = (server_config *)ap_get_module_config(
      cmd->server->module_config, &vhosterror_module);
  TRANS *t = (TRANS *)priorities;
  const char *err = ap_check_cmd_context(cmd,
      NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);

  if (err != NULL)
    return err;
  for(;;)
  {
    if (t->t_name == NULL)
      return apr_psprintf(cmd->pool, "Unknown priority: %s", p1);
    if (strcmp(t->t_name, p1) == 0)
    {
      sconf->loglevel = t->t_val;
      return NULL;
    }
    t++;
  }
}

static const command_rec command_table[] = {
  AP_INIT_TAKE1("VhostErrorLog", cmd_vhost_error_log, NULL, RSRC_CONF,
      "The filename of the vhost error log file"),
  AP_INIT_TAKE1("VhostLogLevel", cmd_vhost_error_log_level, NULL, RSRC_CONF,
      "Vhost error log level. Same as LogLevel but LogLevel precedes"),
  { NULL },
};

static void error_log(const char *file, int line, int level,
    apr_status_t status, const server_rec *s, const request_rec *r,
    apr_pool_t *pool, const char *fmt)
{
  server_config *sconf;
  apr_size_t len = 0;
  char errstr[MAX_STRING_LEN];
  int level_and_mask = level & APLOG_LEVELMASK;

  /* we'are only interested in log output within request context */
  if (s == NULL || r == NULL)
    return;
  sconf = (server_config *)ap_get_module_config(s->module_config, &vhosterror_module);
  if (sconf == NULL || sconf->fd == NULL)
    return;

  if (level_and_mask > sconf->loglevel)
    return;

  /* prepend vhost */
  len += apr_snprintf(errstr + len, MAX_STRING_LEN - len,
      "%s ", ap_escape_logitem(r->pool, r->server->server_hostname));

  /* date */
  errstr[len++] = '[';
  ap_recent_ctime(errstr + len, apr_time_now());
  errstr[len + APR_CTIME_LEN - 1] = ']';
  errstr[len + APR_CTIME_LEN    ] = ' ';
  len += APR_CTIME_LEN + 1;

  /* priority */
  len += apr_snprintf(errstr + len, MAX_STRING_LEN - len,
      "[%s] ", priorities[level_and_mask].t_name);

  /* client ip + log message */
  len += apr_snprintf(errstr + len, MAX_STRING_LEN - len,
      "[client %s] %s", r->connection->remote_ip, fmt);

  apr_file_puts(errstr, sconf->fd);
  apr_file_flush(sconf->fd);
}

static int open_logs(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
  server_config *sconf = (server_config *)ap_get_module_config(s->module_config,
      &vhosterror_module);
  apr_status_t rc;
  piped_log *pl = NULL;

  if (sconf == NULL || sconf->filename == NULL)
    return OK;

  if (sconf->filename[0] == '|')
  {
    if ((pl = ap_open_piped_log(p, sconf->filename + 1)) == NULL)
    {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
          "mod_vhosterror: could not open pipe to %s", sconf->filename);
      return DONE;
    }
    sconf->fd = ap_piped_log_write_fd(pl);
  }
  else
  {
    char *filename = ap_server_root_relative(p, sconf->filename);
    if (filename == NULL)
    {
      ap_log_error(APLOG_MARK, APLOG_ERR, APR_EBADPATH, s,
          "mod_vhosterror: invalid filename %s", sconf->filename);
      return DONE;
    }

    if ((rc = apr_file_open(&sconf->fd, filename,
            APR_WRITE | APR_APPEND | APR_CREATE | APR_LARGEFILE,
            APR_OS_DEFAULT, p)) != APR_SUCCESS)
    {

      ap_log_error(APLOG_MARK, APLOG_STARTUP, rc, s,
          "mod_vhosterror: could not open file %s ", filename);
      return DONE;
    }
  }

  return OK;
}

static void *create_server_config(apr_pool_t *p, server_rec *s)
{
  server_config *sconf = (server_config *)apr_pcalloc(p, sizeof(server_config));
  sconf->filename = NULL;
  sconf->loglevel = APLOG_NOTICE;
  sconf->fd = NULL;
  return sconf;
}

static void register_hooks(apr_pool_t *p)
{
  ap_hook_open_logs(open_logs, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_error_log(error_log, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA vhosterror_module =
{
  STANDARD20_MODULE_STUFF,
  NULL,                           /* per-directory config creator */
  NULL,                           /* dir config merger */
  create_server_config,           /* server config creator */
  NULL,                           /* server config merger */
  command_table,                  /* command table */
  register_hooks,                 /* set up other request processing hooks */
};
